/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "signal/signal.pmh"	// direct access to ChainSignaler

#include "network/network.h"

#define LISTENQ		5
#define SERV_PORT	7876

int main(void)
{
	fe::UnitTest unitTest;

	try
	{
		fe::sp<fe::Socket>		listener(new fe::Socket);
		fe::sp<fe::Socket>		connection(new fe::Socket);
		fe::SockAddr			server_addr;
		fe::SockAddr			client_addr;

		listener->openStream();
		server_addr.setPort(SERV_PORT);
		listener->bind(server_addr);
		listener->listen(LISTENQ);

		fe::sp<fe::Master> spMaster(new fe::Master);
		fe::sp<fe::Scope> spScope(new fe::Scope(*spMaster.raw()));

		fe::sp<fe::SignalerI> spSignalerI(new fe::ChainSignaler());

		fe::Accessor<int>	a(spScope,"a");
		fe::sp<fe::Layout>	spSignal = spScope->declare("SIGNAL");
		spSignal->populate(a);

		connection = listener->accept(client_addr);

		fe::sp<fe::ChunkSpool> spChunkSpool(new fe::ChunkSpool);

		fe::sp<fe::ChunkSender> spChunkSender(new fe::ChunkSender);

		fe::sp<fe::HandlerI> spHandlerI(new fe::SignalSender(
			connection->id(), spScope, spChunkSpool));
		spSignalerI->insert(spHandlerI, spSignal);

		fe::Record signal = spScope->createRecord(spSignal);
		a(signal) = 42;
		spSignalerI->signal(signal);

		spChunkSender->start(spChunkSpool, connection);

		while(true)
		{
			fe::milliSleep(500);
			spSignalerI->signal(signal);
		}

		spChunkSender->stop();

	}
	catch(fe::Exception &e)
	{
		e.log();
	}
	catch(...)
	{
		feLogGroup("network","uncaught exception");
	}

	return unitTest.failures();
}
