/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "viewer.pmh"

#define FE_OV_DISPLAY_FRAMERATE		TRUE

namespace fe
{

#define FRAMECOUNT_PERIOD	500.0f	//* ms

ObjectViewer::ObjectViewer(void)
{
	m_frameMS=0.0f;
	m_frameRate=0.0f;
	m_frameCount=0;
	m_frameTotal=0.0f;
}

ObjectViewer::~ObjectViewer(void)
{
#ifdef FE_CHAINSIGNALER_PROFILING
	Peeker peeker;
	peek(peeker);

	feLog("~ObjectViewer\n%s\n",peeker.str().c_str());
#endif
}

void ObjectViewer::initialize(void)
{
	m_spControllerI=registry()->create("ControllerI");
	if(!m_spControllerI.isValid())
		feX("ObjectViewer::initialize","needed to create an ControllerI");
}

void ObjectViewer::bind(sp<WindowI> spWindowI)
{
	m_spDrawI->view()->setWindow(spWindowI);

	//* TODO should use signaler/scope from WindowI, not our signaler bind
//	sp<EventContextI> spEventContextI=spWindowI->getEventContextI();
}

void ObjectViewer::bind(sp<SignalerI> spSignalerI,sp<Layout> spLayout)
{
	FEASSERT(spSignalerI.isValid());

	sp<Scope> spScope=spLayout->scope();
	FEASSERT(spScope.isValid());

	if(m_spScope!=spScope)
	{
		if(m_spScope.isValid())
			feLog("%p ObjectViewer::bind() changed scope\n",this);

		m_spScope=spScope;

		sp<Layout> spEventLayout=spScope->lookupLayout(FE_EVENT_LAYOUT);
		FEASSERT(spEventLayout.isValid());

		sp<HandlerI> spHandlerI=m_spControllerI;
		FEASSERT(spHandlerI.isValid());

		spSignalerI->insert(spHandlerI,spEventLayout);
	}

	m_asViewer.bind(spScope);
}

void ObjectViewer::handle(Record &r_sig)
{
	const Real nearplane=0.1f;
	const Real farplane=1e4;
	const Real fovy=30.0f;

	sp<ViewI> spViewI=m_spDrawI->view();
	sp<ViewMode> spViewMode=spViewI->viewMode();
	sp<WindowI> spWindowI=spViewI->window();
	Box2i viewport=spWindowI->geometry();
	viewport[0]=0;
	viewport[1]=0;

	I32 layer= -1;
	if(m_asViewer.viewer_layer.check(r_sig))
	{
		layer=m_asViewer.viewer_layer(r_sig);
	}

	if(layer==0)
	{
		const Real ms=m_ticker.timeMS();
		m_frameTotal+=ms-m_frameMS;
		m_frameCount++;
		if(m_frameTotal>FRAMECOUNT_PERIOD)
		{
			m_frameRate=m_frameCount*1e3f/m_frameTotal;
			m_frameTotal=0.0f;
			m_frameCount=0;
		}
		m_frameMS=ms;
	}

	const BWORD perspective=(m_asViewer.viewer_perspective.check(r_sig) &&
			m_asViewer.viewer_perspective(r_sig));

	if(perspective)
	{
		spViewMode->setCamera(m_spControllerI->cameraMatrix());
		spViewMode->setPerspective(fovy,nearplane,farplane);
		spViewMode->setViewport(viewport);
		spViewI->use(ViewMode::e_perspective);
	}
	else
	{
		spViewMode->setViewport(viewport);
		spViewI->use(ViewMode::e_ortho);

#if FE_OV_DISPLAY_FRAMERATE
		if(layer>0)
		{
			const Color pink(1.0f,0.5f,0.5f);
			const SpatialVector corner(10.0f,10.0f,0.0f);

			String text;
			text.sPrintf("%.0f FPS",frameRate());
			m_spDrawI->drawAlignedText(corner,text,pink);
		}
#endif
	}

	signal(r_sig);
}

} /* namespace */
