/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "xCatalog.h"

using namespace fe;

int main(int argc,char** argv)
{
	UNIT_START();

	BWORD complete=FALSE;

	try
	{
		sp<Master> spMaster(new Master);
		sp<Registry> spRegistry=spMaster->registry();

		Result result=spRegistry->manage("xCatalogLibrary");
		UNIT_TEST(successful(result));

		sp<Catalog> spCatalog(spRegistry->create("Catalog"));
		UNIT_TEST(spCatalog.isValid());
		if(!spCatalog.isValid())
		{
			feX(argv[0], "NULL Catalog");
		}

		spCatalog->catalog<String>("a string")="string value";

/*
		spCatalog->catalogSet("a real", (Arg[]){
				{1.2f,"min"},
				3.4f,
				{5.6f,"max"},
				{"text","label"},
				(void*)NULL });
*/

		spCatalog->catalog<I32>("integer")= -7;
		spCatalog->catalog<String>("integer","label")="An Integer";
		spCatalog->catalog<I32>("integer","min")=2;
		spCatalog->catalog<I32>("integer","max")=8;

		I32 value=spCatalog->catalog<I32>("integer");
		printf("value %d\n", value);
		UNIT_TEST(value== -7);

		spCatalog->catalogDump();

		spCatalog->catalog<Real>("float")= 1.23;
		spCatalog->catalog<String>("float","label")="A Float";

		spCatalog->catalogDump();

		value=spCatalog->catalogOrDefault<I32>("integer",13);
		printf("value %d\n", value);
		UNIT_TEST(value== -7);

		spCatalog->catalogDump();

		spCatalog->catalogRemove("integer");
		UNIT_TEST(!spCatalog->cataloged("integer"));

		spCatalog->catalogDump();

		value=spCatalog->catalogOrDefault<I32>("integer",13);
		printf("value %d\n", value);
		UNIT_TEST(value==13);
		UNIT_TEST(!spCatalog->cataloged("integer"));

		spCatalog->catalogDump();

		UNIT_TEST(spMaster->catalog().isValid());

		complete=TRUE;
	}
	catch(Exception &e) { e.log(); }
	catch(...)			{ feLog("uncaught exception\n"); }

	UNIT_TEST(complete);
	UNIT_TEST(!Counted::trackerCount());
	UNIT_TRACK(6);
	UNIT_RETURN();
}
